'''
Author: Pengyao PING
Date: 2022-06-20 15:56:46
LastEditors: Pengyao PING
LastEditTime: 2022-10-21 13:57:29
Email: Pengyao.Ping@student.uts.edu.au
Description: 
'''
from SRRec.config import Config
import sys, getopt
from SRRec.data_generation import DataGneration
from SRRec.error_orrection import ErrorCorrection
from SRRec.data_analysis import DataAnalysis
import os
from SRRec.simulation import Simulation
from SRRec.utils import custom_logger, usage
from SRRec.data_preprocessing import DataProcessing

def main():
    argv = sys.argv[1:]
    # create logger
    logger = custom_logger("SRRec", debug_mode=True)
    ##############################################################
    try:
        opts, args = getopt.getopt(argv, "m:c:i:u:t:r:d:p:a:g:h:", ["module=", "config=", "input=", "umi_file", "true", "rectification", "directory", "parallel", "high_ambiguous", "tree_method", "help"]) 
    except getopt.GetoptError as err:
        if not ('-h' or "--help" in opts):
            logger.error(err)
        usage()
        raise
    required_ags = False
    # print(opts)
    if opts:
        opts_dict = dict(opts)
        opts_keys = list(opts_dict.keys())
        tar_set = set(opts_keys)
        # print(tar_set)
        h_lst = list({'-h', '--help'}.intersection(tar_set))
        if h_lst:
            usage()                
            required_ags = True 
            sys.exit()  
        else: 
            m_lst = list({"-m", "--module"}.intersection(tar_set))
            c_lst = list({"-c", "--config"}.intersection(tar_set)) 
            i_lst = list({"-i", "--input"}.intersection(tar_set)) 
            u_lst = list({"-u", "--umi_file"}.intersection(tar_set)) 
            t_lst = list({"-t", "--true"}.intersection(tar_set)) 
            r_lst = list({"-r", "--rectification"}.intersection(tar_set))   
            d_lst = list({"-d", "--directory"}.intersection(tar_set))
            p_lst = list({"-p", "--parallel"}.intersection(tar_set))
            a_lst = list({"-a", "--high_ambiguous"}.intersection(tar_set))
            g_lst = list({"-g", "--tree_method"}.intersection(tar_set))
            try:
                module_arg = opts_dict[m_lst[0]]
                #############################################################
                if module_arg == "correction": 
                    if c_lst:
                        config = Config(opts_dict[c_lst[0]], logger) 
                        required_ags = True   
                        if i_lst:
                            config.input_file = opts_dict[i_lst[0]]
                    elif i_lst:
                        config = Config(None, logger)  
                        config.input_file = opts_dict[i_lst[0]]
                        required_ags = True
                    if t_lst:
                        config.ground_truth_dataset = opts_dict[t_lst[0]] 
                    if d_lst:
                        config.ResultDir = opts_dict[d_lst[0]] 
                    if p_lst:
                        config.num_workers = int(opts_dict[p_lst[0]])      
                    if a_lst:
                        config.high_ambiguous = opts_dict[a_lst[0]]
                    if g_lst:
                        config.tree_method = opts_dict[g_lst[0]]
                #############################################################
                elif module_arg == "amplicon_correction": 
                    if c_lst:
                        config = Config(opts_dict[c_lst[0]], logger) 
                        required_ags = True   
                        if i_lst:
                            config.input_file = opts_dict[i_lst[0]]
                    elif i_lst:
                        config = Config(None, logger)  
                        config.input_file = opts_dict[i_lst[0]]
                        required_ags = True
                    if t_lst:
                        config.ground_truth_dataset = opts_dict[t_lst[0]] 
                    if d_lst:
                        config.ResultDir = opts_dict[d_lst[0]] 
                    if p_lst:
                        config.num_workers = int(opts_dict[p_lst[0]])    
                    if a_lst:
                        config.high_ambiguous = opts_dict[a_lst[0]] 
                    if g_lst:
                        config.tree_method = opts_dict[g_lst[0]] 
                #############################################################
                elif module_arg == "umi_correction": 
                    if c_lst:
                        config = Config(opts_dict[c_lst[0]], logger) 
                        required_ags = True   
                        if i_lst:
                            config.input_file = opts_dict[i_lst[0]]
                    elif i_lst:
                        config = Config(None, logger)  
                        config.input_file = opts_dict[i_lst[0]]
                        required_ags = True
                    if t_lst:
                        config.ground_truth_dataset = opts_dict[t_lst[0]] 
                    if d_lst:
                        config.ResultDir = opts_dict[d_lst[0]] 
                    if p_lst:
                        config.num_workers = int(opts_dict[p_lst[0]])  
                    if a_lst:
                        config.high_ambiguous = opts_dict[a_lst[0]] 
                #############################################################   
                elif module_arg == "mimic_umi":   
                    if i_lst and t_lst:
                        config = Config(None, logger)  
                        config.input_file = opts_dict[i_lst[0]]
                        required_ags = True
                        config.ground_truth_dataset = opts_dict[t_lst[0]] 
                    else:
                        logger.error("Must input raw and true datasets simultaneously.")
                    if d_lst:
                        config.ResultDir = opts_dict[d_lst[0]] 
                #############################################################
                elif module_arg == "real_umi":   
                    if i_lst:
                        config = Config(None, logger)  
                        config.input_file = opts_dict[i_lst[0]]
                        required_ags = True
                    else:
                        logger.error("Must input raw and true datasets simultaneously.")
                    if u_lst:
                        config.umi_file = opts_dict[u_lst[0]] 
                    if d_lst:
                        config.ResultDir = opts_dict[d_lst[0]] 
                #############################################################
                elif module_arg == "evaluation":   
                    if i_lst and t_lst and r_lst:
                        config = Config(None, logger)  
                        config.input_file = opts_dict[i_lst[0]]
                        config.ground_truth_dataset = opts_dict[t_lst[0]]
                        config.rectified_data = opts_dict[r_lst[0]]
                        required_ags = True 
                    else:
                        logger.error("Must input raw, true and rectified datasets simultaneously.")
                    if d_lst:
                        config.ResultDir = opts_dict[d_lst[0]] 
                #############################################################
                elif module_arg == "simulation":   
                    if i_lst:
                        config = Config(None, logger)  
                        config.input_file = opts_dict[i_lst[0]]
                        required_ags = True 
                    else:
                        logger.error("Must input a raw dataset.")
                    if d_lst:
                        config.ResultDir = opts_dict[d_lst[0]] 
                else:
                    logger.exception("Please check the module name.")
                    sys.exit()
                available_cpu_cores = os.cpu_count()
                if config.num_workers <= 0:
                    config.num_workers = available_cpu_cores - 2
                if config.num_workers > available_cpu_cores:
                    logger.error(f"Only {available_cpu_cores} available to use.") 
                    config.num_workers = available_cpu_cores - 2
            except:
                logger.exception("Not or incorrectly select module (e.g. correction, evaluation, mimic_umi, simulation)")
                raise
        if not required_ags:
            logger.exception("Must select module and input config file (*.ini) or read file (e.g. *.fastq).")
            sys.exit()
    else:
        logger.exception("No options supplied!")
        raise

    if module_arg == "correction":
        ##############################################################
        DG = DataGneration(
            logger,
            config.num_workers,
            config.input_file, 
            config.ResultDir,
            config.high_freq_thre,
            config.max_error_freq,
            config.save_graph,
            config.graph_visualization,
            config.drawing_graph_num,
            config.high_ambiguous,
            config.min_iters,
            config.verbose)
        if config.high_ambiguous:
            isolates_file, non_isolates_file, unique_seqs, read_max_len, read_min_len, genuine_df, negative_df, ambiguous_df, high_ambiguous_df = DG.data_files(edit_dis=1)
        else:
            isolates_file, non_isolates_file, unique_seqs, read_max_len, read_min_len, genuine_df, negative_df, ambiguous_df = DG.data_files(edit_dis=1)      
        # read_max_len=238     
        ###############################################################
        EC = ErrorCorrection(
            logger,
            config.num_workers,
            config.input_file, 
            config.ResultDir,
            read_max_len,
            config.read_type)
        EC.ambiguous_paras(
            config.n_trials,
            config.test_size, 
            config.random_state, 
            config.tree_method,
            config.learning_rate_min, 
            config.learning_rate_max, 
            config.max_depth_min, 
            config.max_depth_max, 
            config.max_depth_step, 
            config.num_boost_round_min, 
            config.num_boost_round_max, 
            config.num_boost_round_step, 
            config.subsample_min, 
            config.subsample_max, 
            config.colsample_bytree_min, 
            config.colsample_bytree_max, 
            config.verbose_eval,
            config.seed,
            config.best_accuracy,
            config.n_estimators,
            config.proba_deviation)
        ############################################################################################
        ## one model to predict
        if config.high_ambiguous:
            ##################################
            corrected_file, no_high_correct, new_negative_df = EC.all_in_one_correction(isolates_file, non_isolates_file, unique_seqs, genuine_df, negative_df, ambiguous_df, high_ambiguous_df)
        else:
            corrected_file, new_negative_df = EC.all_in_one_correction(isolates_file, non_isolates_file, unique_seqs, genuine_df, negative_df, ambiguous_df, high_ambiguous_df =None) 

        ########################################################################################### 
        # for debuging ed2 correction 
        # corrected_file = os.path.join(config.ResultDir, "non_umi_SRR1543964_corrected.fastq")         
        # genuine_csv2 = os.path.join(config.ResultDir, "genuine2.csv")
        # ambiguous_csv2 = os.path.join(config.ResultDir, "ambiguous2.csv")
        # high_ambiguous_csv2 = os.path.join(config.ResultDir, "high_ambiguous2.csv")
        # negative_csv2 = os.path.join(config.ResultDir, "negative2.csv")
        # read_min_len = 100
        
        ###########
        if read_min_len > 30:
            genuine_df2, negative_df2, ambiguous_df2, unique_seqs2 = DG.extract_ed2_errors(corrected_file)
            final_result = EC.all_in_one_ed2_correction(corrected_file, unique_seqs2, genuine_df2, negative_df2, ambiguous_df2)
        else:
            final_result = corrected_file
        DA = DataAnalysis(
            logger,
            config.num_workers,
            config.input_file, 
            final_result,
            config.ground_truth_dataset,
            config.ResultDir, 
            top_n = 100)
        DA.evaluation()  
    ############################################
        if config.high_ambiguous:
            if read_min_len > 30:
                genuine_df3, negative_df3, ambiguous_df3, unique_seqs3 = DG.extract_ed2_errors(no_high_correct)
                final_result_no_high = EC.all_in_one_ed2_correction(no_high_correct, unique_seqs3, genuine_df3, negative_df3, ambiguous_df3)
            else:
                final_result_no_high = no_high_correct
            DA = DataAnalysis(
                logger,
                config.num_workers,
                config.input_file, 
                final_result_no_high,
                config.ground_truth_dataset,
                config.ResultDir, 
                top_n = 100)
            DA.evaluation()  

    elif module_arg == "amplicon_correction":
        ##############################################################
        DG = DataGneration(
            logger,
            config.num_workers,
            config.input_file, 
            config.ResultDir,
            config.high_freq_thre,
            config.max_error_freq,
            config.save_graph,
            config.graph_visualization,
            config.drawing_graph_num,
            config.high_ambiguous,
            config.min_iters,
            config.verbose
            )      
        if config.high_ambiguous:
            isolates_file, non_isolates_file, unique_seqs, read_max_len, read_min_len, genuine_df, negative_df, ambiguous_df, high_ambiguous_df = DG.data_files(edit_dis=1)
        else:
            isolates_file, non_isolates_file, unique_seqs, read_max_len, read_min_len, genuine_df, negative_df, ambiguous_df = DG.data_files(edit_dis=1)     
        ###############################################################
        EC = ErrorCorrection(
            logger,
            config.num_workers,
            config.input_file, 
            config.ResultDir,
            read_max_len,
            config.read_type)
        EC.ambiguous_paras(
            config.n_trials,
            config.test_size, 
            config.random_state, 
            config.tree_method,
            config.learning_rate_min, 
            config.learning_rate_max, 
            config.max_depth_min, 
            config.max_depth_max, 
            config.max_depth_step, 
            config.num_boost_round_min, 
            config.num_boost_round_max, 
            config.num_boost_round_step, 
            config.subsample_min, 
            config.subsample_max, 
            config.colsample_bytree_min, 
            config.colsample_bytree_max, 
            config.verbose_eval,
            config.seed,
            config.best_accuracy,
            config.n_estimators,
            config.proba_deviation)
        if config.high_ambiguous:
            corrected_file, no_high_correct, new_negative_df = EC.all_in_one_correction(isolates_file, non_isolates_file, unique_seqs, genuine_df, negative_df, ambiguous_df, high_ambiguous_df)
        else:
            corrected_file, new_negative_df = EC.all_in_one_correction(isolates_file, non_isolates_file, unique_seqs, genuine_df, negative_df, ambiguous_df, high_ambiguous_df = None) 

        if read_min_len > 30:
            genuine_df2, negative_df2, ambiguous_df2, unique_seqs2 = DG.extract_ed2_errors(corrected_file)
            mid_result = EC.all_in_one_ed2_correction(corrected_file, unique_seqs2, genuine_df2, negative_df2, ambiguous_df2,)
        else:
            mid_result = corrected_file

        amplicon_df = DG.extract_amplicon_err_samples(mid_result, config.amplicon_low_freq, config.amplicon_high_freq)
        final_result = EC.correct_amplicon_err(mid_result, genuine_df, negative_df, new_negative_df, amplicon_df, config.amplicon_threshold_proba)
        #####
        DA = DataAnalysis(
            logger,
            config.num_workers,
            config.input_file, 
            final_result,
            config.ground_truth_dataset,
            config.ResultDir, 
            top_n = 100)
        DA.evaluation() 
         
    elif module_arg == "umi_correction":
        ##############################################################
        high_ambiguous=False
        DG = DataGneration(
            logger,
            config.num_workers,
            config.input_file, 
            config.ResultDir,
            config.high_freq_thre,
            config.max_error_freq,
            config.save_graph,
            config.graph_visualization,
            config.drawing_graph_num,
            high_ambiguous,
            config.min_iters,
            config.verbose
            )
        genuine_df = DG.extract_umi_genuine_errs()
        # ##############################################################
        EC = ErrorCorrection(
            logger,
            config.num_workers,
            config.input_file, 
            config.ResultDir,
            read_max_len,
            config.read_type)
        corrected_file = EC.umi_correction(config.input_file, genuine_df)
        
        DA = DataAnalysis(
            logger,
            config.num_workers,
            config.input_file, 
            corrected_file,
            config.ground_truth_dataset,
            config.ResultDir, 
            top_n = 100)
        DA.evaluation()  

    elif module_arg == "mimic_umi":
        DP = DataProcessing( 
            logger,
            config.num_workers,
            config.ResultDir)
        DP.write_mimic_umis(config.input_file, config.ground_truth_dataset)
        
    elif module_arg == "real_umi":
        DP = DataProcessing( 
            logger,
            config.num_workers,
            config.ResultDir,
            config.umi_start,
            config.umi_end,
            config.non_umi_start)
        DP.real_umi_data(config.input_file)        

    ##############################################################
    elif module_arg == "evaluation":
        DA = DataAnalysis(
            logger,
            config.num_workers,
            config.input_file, 
            config.rectified_data,
            config.ground_truth_dataset,
            config.ResultDir, 
            top_n = 100)
        DA.evaluation()  

    ##############################################################
    elif module_arg == "simulation":
        Sim = Simulation(
            logger,
            config.num_workers,
            config.input_file, 
            config.error_rate,
            config.ResultDir,
            config.substations,
            config.indels,
            config.seed
            )
        Sim.simulation()
# if __name__ == '__main__':
#     main(sys.argv[1:])